/*
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
 * 
 * Contact Information:
 * Dambach-Werke GmbH
 * Elektronische Leitsysteme
 * Fritz-Minhardt-Str. 1
 * 76456 Kuppenheim
 * Phone: +49-7222-402-0
 * Fax: +49-7222-402-200
 * mailto: info@els.dambach.de
 */

package de.bsvrz.vew.sim.sim.datengenerator;

import java.util.ArrayList;
import java.util.Collection;
import java.util.HashMap;
import java.util.HashSet;
import java.util.Map;

import de.bsvrz.dav.daf.main.ClientDavInterface;
import de.bsvrz.dav.daf.main.ClientSenderInterface;
import de.bsvrz.dav.daf.main.Data;
import de.bsvrz.dav.daf.main.DataDescription;
import de.bsvrz.dav.daf.main.ResultData;
import de.bsvrz.dav.daf.main.SendSubscriptionNotConfirmed;
import de.bsvrz.dav.daf.main.Data.Array;
import de.bsvrz.dav.daf.main.config.Aspect;
import de.bsvrz.dav.daf.main.config.AttributeGroup;
import de.bsvrz.dav.daf.main.config.AttributeGroupUsage;
import de.bsvrz.dav.daf.main.config.ConfigurationArea;
import de.bsvrz.dav.daf.main.config.ObjectSet;
import de.bsvrz.dav.daf.main.config.SystemObject;
import de.bsvrz.dav.daf.main.config.SystemObjectType;
import de.bsvrz.dav.daf.main.config.AttributeGroupUsage.Usage;
import de.bsvrz.sys.funclib.debug.Debug;
import de.bsvrz.vew.sim.sim.simulation.Simulation;
import de.bsvrz.vew.sim.sim.simulationsstrecke.SimulationsStrecke;
  

/**
 * Datengenerator für Simulationen. Funktionalität zum analysieren der Eingangsdaten
 * und zum Senden der Ausgangsdaten
 *
 * @version $Revision: 1.3 $ / $Date: 2008/10/31 13:30:31 $ / ($Author: Sans $)
 *
 * @author Dambach-Werke GmbH
 * @author Stefan Sans
 *
 */
public abstract class DatenGeneratorSimulation implements ClientSenderInterface
{
  /**
   * Debug-Logger für Logging-Ausgaben
   */
  private final static Debug _debug = Debug.getLogger();

  /**
   * Anzahl der momentan aktiven Sende-Threads
   */
  private int _anzahlSendeThreads = 0;

  /**
   * Die Simulation, für die diese Instanz Daten generiert
   */
  private Simulation _simulation;


  public DatenGeneratorSimulation(Simulation simulation)
  {
    _simulation = simulation;
  }

  /**
   * Bestimmt den Lauf-Status des Datengenerators
   * @return Ist der Datengenerator gestartet?
   */
  public abstract boolean isGestartet();

  /**
   * Initilaisiere den Datengenrator
   */
  protected abstract void init();

  /**
   * Starte den Datengenerator
   */
  public abstract void start();

  /*
   * Stoppe den Datengenerator
   */
  public abstract void stop();

  /**
   * Bestimmt die Simulation, für die diese Instanz Daten generiert
   * @return die Simulation, für die diese Instanz Daten generiert
   */
  protected Simulation getSimulation()
  {
    return _simulation;
  }

  /**
   * Bestimmt die Simulationstrecke der Simulation, für die diese Instanz
   * Daten generiert
   * @return die Simulationsstrecke der Simulation, für die diese Instanz
   * Daten generiert
   */
  protected SimulationsStrecke getSimulationsStrecke()
  {
    return _simulation.getSimulationsStrecke();
  }

  /**
   * Bestimmt die Datenverteilerverbindung
   * @return die Datenverteilerverbindung
   */
  protected ClientDavInterface getConnection()
  {
    return _simulation.getConnection();
  }

  /**
   * Bestimmt die zu betrachtenden Konfigurationsbereiche aus den Eingangsdaten
   * @param bereiche Feld mit den Bereichs-Spezifikationen
   * @param index Index des Parametersatzes
   * @return die Bereiche als Sammlung von (Stellvertreter-)Systemobjekten
   */
  private Collection<SystemObject> getBereiche(Array bereiche, int index)
  {
    Collection<SystemObject> col;

    if (bereiche != null && bereiche.getLength() > 0)
    {
      col = new ArrayList<SystemObject>(bereiche.getLength());
      for (int i = 0; i < bereiche.getLength(); i += 1)
      {
       col.add(bereiche.getItem(i).asReferenceValue().getSystemObject());
      }
      _debug.fine(col.size() + " Konfigurationsbereich(e) werden bearbeitet für " +
          (index + 1)  + ". Parametersatz der Simulationsstrecke " + getSimulationsStrecke());
    }
    else
    {
      SystemObjectType type = getConnection().getDataModel().getType("typ.konfigurationsBereich");
      col = type.getObjects();
      _debug.fine("Keine Bereiche angegeben für " + (index + 1)  +
          ". Parametersatz der Simulationsstrecke " + getSimulationsStrecke() +
          "\n=> alle möglichen " + col.size() +
          "(!) Konfigurationsbereiche werden bearbeitet");
    }

    return col;
  }

  /**
   * Bestimmt die zu betrachtetenden Objekte aus den Eingangsdaten
   * @param objekte Feld mit Objekten-Spezifikationen aus den Eingangsdaten
   * @param bereiche Sammmlung der Bereiche, aus denen die Objekte zulässig sind
   * @param datenIndex Daten-Index innerhalb des Parametersatzes
   * @param parameterIndex Index des Parametersatzes
   * @return die Objekte als Sammlung von (Stellvertreter-)Systemobjekten
   */
  private Collection<SystemObject> getObjekte(Array objekte, Collection<SystemObject> bereiche,
      int datenIndex, int parameterIndex)
  {
    Collection<SystemObject> col = new ArrayList<SystemObject>();

    _debug.finer("Hole Objekte gemäss " + (datenIndex + 1) + ". Datenblock des " +
        (parameterIndex + 1) + ". Parametersatzes der Simulationsstrecke " + getSimulationsStrecke());

    if (objekte != null && objekte.getLength() > 0)
    {
      for (int i = 0; i < objekte.getLength(); i += 1)
      {
        SystemObject obj = objekte.getItem(i).asReferenceValue().getSystemObject();
        Collection<SystemObject> toAdd = null;

        if (obj.isOfType("typ.typ"))
        {
          toAdd = ((SystemObjectType)obj).getObjects();
        }
        else if (obj.isOfType("typ.menge"))
        {
          toAdd = ((ObjectSet)obj).getElements();
        }

        for (SystemObject el: toAdd == null ? new SystemObject[] {obj} :
          toAdd.toArray(new SystemObject[toAdd.size()]))
        {
          if (bereiche.contains(el.getConfigurationArea()))
          {
            col.add(el);
          }
          else
          {
            _debug.warning("Verwerfe Objekt " + el + " - wegen inkonsistentem Konfigurationsbereich - im " +
                (datenIndex + 1) + ". Datenblock des " + (parameterIndex + 1)  +
                ". Parametersatzes der Simulationsstrecke " + getSimulationsStrecke());
          }
        }
      }
      _debug.fine(col.size() + " Objekte werden bearbeitet für " + (datenIndex + 1) +
          ". Datenblock des " + (parameterIndex + 1) + ". Parametersatzes der Simulationsstrecke " +
          getSimulationsStrecke());
    }
    else
    {
      for (SystemObject obj: bereiche)
      {
        col.addAll(((ConfigurationArea)obj).getCurrentObjects());
      }

      _debug.fine("Keine Objekte angegeben für " + (datenIndex + 1) + ". Datenblock des " +
          (parameterIndex + 1)  + ". Parametersatzes der Simulationsstrecke " + getSimulationsStrecke() +
          "\n=> " + col.size() + " Objekte aus allen Bereichen (" + bereiche.size() + ") werden bearbeitet");
    }

    return col;

  }

  /**
   * Bestimmt die zu betrachtenden Attributgruppen aus den Eingangsdaten
   * @param attributGruppen Feld mit Attributgruppen-Spezifikationen
   * @return Sammlung von Attributgruppen
   */
  private Collection<AttributeGroup> getAttributGruppen(Array attributGruppen)
  {
    Collection<AttributeGroup> atgn = new HashSet<AttributeGroup>();

    if (attributGruppen != null)
    {
      for (int i = 0 ; i < attributGruppen.getLength(); i += 1)
      {
        atgn.add((AttributeGroup)attributGruppen.getItem(i).asReferenceValue().getSystemObject());
      }
    }

    return atgn;
  }

  /**
   * Bestimmt die zu betrachtenden Aspekte aus den Eingangsdaten
   * @param aspekte Feld mit Aspekt-Spezifikationen
   * @return Sammlung von Aspekten
   */
  private Collection<Aspect> getAspekte(Array aspekte)
  {
    Collection<Aspect> aspe = new HashSet<Aspect>();

    if (aspekte != null)
    {
      for (int i = 0 ; i < aspekte.getLength(); i += 1)
      {
        aspe.add((Aspect)aspekte.getItem(i).asReferenceValue().getSystemObject());
      }
    }

    return aspe;
  }

  /**
   * Ein Block von Eingangsdaten wird analysiert
   * @param daten Datensatz des Eingangsdaten-Blocks
   * @param bereiche Sammlung der zulässigen Bereiche
   * @param eingangsDaten die analysierten Eingangsdaten werden hier hinzugefügt
   * @param datenIndex Daten-Index innerhalb Parametersatz
   * @param parameterIndex Indes des Parametrer-Satzes
   */
  private void analysiereDatenBlock(Data daten, Collection<SystemObject> bereiche,
      Map<String, DatenSpezifikation> eingangsDaten, int datenIndex, int parameterIndex)
  {
    Array array = daten.getArray("Objekt");
    Collection<SystemObject> objekte = getObjekte(array, bereiche, datenIndex, parameterIndex);

    array = daten.getArray("AttributGruppe");
    Collection<AttributeGroup> attributGruppen = getAttributGruppen(array);

    array = daten.getArray("Aspekt");
    Collection<Aspect> aspekte = getAspekte(array);

    for (SystemObject obj: objekte)
    {
      SystemObjectType typ = obj.getType();

      Collection<AttributeGroup> atgn = typ.getAttributeGroups();

      for (AttributeGroup atg: atgn)
      {
        if (attributGruppen.isEmpty() || attributGruppen.contains(atg))
        {
          Collection<Aspect> aspe = atg.getAspects();

          for (Aspect asp: aspe)
          {
            // Für parametrierende Daten ist nur der Aspekt asp.parameterSoll zugelassen
            if (!atg.isParameter() || asp.getPid().equals("asp.parameterSoll"))
            {
              if (aspekte.isEmpty() || aspekte.contains(asp))
              {
                AttributeGroupUsage atgv = atg.getAttributeGroupUsage(asp);
                Usage benutzung = atgv.getUsage();
                                                                    // Nicht zugelassen sind weiterhin:
                if (!atgv.isConfigurating() &&                      // Konfigurierende Daten
                    !(benutzung == Usage.OnlineDataAsSenderDrain))  // Online Daten, die ausschliesslich als Senke empfangen werden können
                {
                  String key = DatenSpezifikation.erzeugeSchluessel(atg, asp);
                  DatenSpezifikation ds = eingangsDaten.get(key);
                  if (ds == null)
                  {
                    ds = new DatenSpezifikation(atg, asp, (short)0);
                    eingangsDaten.put(key, ds);
                  }
                  ds.getSystemObjekte().add(obj);
                }
              }
            }
          }
        }
      }
    }
  }

  /**
   * Analysiert die Eingangsdaten und gibt das Ergebnis aufbereitet zurück
   * @return Sammlung von Datenspezifikationen als aufbereitetes Eingangsdaten
   */
  protected Collection<DatenSpezifikation> analysiereEingangsDaten()
  {
    _debug.fine("Übernehme Parameter Bereiche/Eingangsdaten von Simulationsstrecke " +
        getSimulationsStrecke() + " für Datengenerator der Online-Simulation " +
        getSimulation());

    Map<String, DatenSpezifikation> eingangsDaten = new HashMap<String, DatenSpezifikation>();
    Array parameterSatz = _simulation.getSimulationsStrecke().getParameterSatz();

    if (parameterSatz != null && parameterSatz.getLength() > 0)
    {
      for (int i = 0; i < parameterSatz.getLength(); i += 1)
      {
        Array array = parameterSatz.getItem(i).getArray("Bereich");
        Collection<SystemObject> bereiche = getBereiche(array, i);

        array = parameterSatz.getItem(i).getArray("Eingangsdaten");
        if (array != null && array.getLength() > 0)
        {
          for (int j = 0; j < array.getLength(); j += 1)
          {
            analysiereDatenBlock(array.getItem(j), bereiche, eingangsDaten, j, i);
          }
        }
        else
        {
          _debug.warning("Keine Eingangsdaten angegeben im " + (i + 1) +
              ". Parametersatz der Simulationsstrecke " + getSimulationsStrecke());
        }
      }
      if (eingangsDaten.isEmpty())
      {
        _debug.warning("Kein Parametersatz für Simulationsdaten in Simulationsstrecke " +
            getSimulationsStrecke() + " generiert nutzbare Eingangsdaten\n=> Simulation " + getSimulation() +
            " wird keinerlei Simulationsdaten publizieren");
      }
    }
    else
    {
      _debug.warning("Keine Parametersätze für Simulationsdaten in Simulationsstrecke " +
          getSimulationsStrecke() + " angegeben\n=> Simulation " + getSimulation() +
          " wird keinerlei Simulationsdaten publizieren");
    }
    return eingangsDaten.values();
  }

  /**
   * Sendet Daten asynchron an den Datenverteiler
   * @param daten die zu sendenden Daten
   */
  protected void sendeDaten(Collection<ResultData> daten)
  {
    if (daten != null && daten.size() > 0)
    {
      ResultData[] data = daten.toArray(new ResultData[daten.size()]);
      sendeDaten(data);
    }
  }

  /**
   * Sendet Daten asynchron an den Datenverteiler
   * @param daten die zu sendenden Daten
   */
  protected void sendeDaten(ResultData[] daten)
  {
    if (daten != null && daten.length > 0)
    {
      final ResultData[] data = daten;

      Thread arbeiter = new Thread()
      {
        public void run()
        {
          try
          {
            _anzahlSendeThreads++;
            _debug.finer(_anzahlSendeThreads + "Sende-Threads laufen für Datengenerator Simulation "  +
                getSimulation());
//            _debug.finer("Datengenerator Simulation "  + getSimulation() +
//                " sendet " + data.length + " Datensa(e)tz(e)");

            String text = "Datengenerator Simulation "  + getSimulation() +
              " sendet " + data.length + " Datensa(e)tz(e)";

            for (int i = 0; i < data.length; i += 1)
            {
              text += "\n\t" +
                data[i].getDataTime() + ":[" + data[i].getObject().getPid() + "]" +
                data[i].getDataDescription().getAttributeGroup().getPid() + "/" +
                data[i].getDataDescription().getAspect()+ "/" +
                data[i].getDataDescription().getSimulationVariant();
            }
            _debug.finer(text);

            getConnection().sendData(data);
          }
          catch (SendSubscriptionNotConfirmed e)
          {
            // Keine Sendesteuerung
            _debug.fine("Keine Sendesteuerung für Datengenerator Simulation "  +
                getSimulation());
          }
          finally
          {
            _anzahlSendeThreads--;
          }
        }
      };
      arbeiter.start();
    }
  }

  public void dataRequest(SystemObject object, DataDescription dataDescription, byte state)
  {
  }

  public boolean isRequestSupported(SystemObject object, DataDescription dataDescription)
  {
    return true;
  }
}
