/*
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
 * 
 * Contact Information:
 * Dambach-Werke GmbH
 * Elektronische Leitsysteme
 * Fritz-Minhardt-Str. 1
 * 76456 Kuppenheim
 * Phone: +49-7222-402-0
 * Fax: +49-7222-402-200
 * mailto: info@els.dambach.de
 */

package de.bsvrz.vew.sim.sim.verwaltung;

import java.text.DecimalFormat;
import java.util.ArrayList;
import java.util.Collection;
import java.util.Collections;
import java.util.HashMap;
import java.util.HashSet;
import java.util.LinkedHashSet;
import java.util.Map;
import java.util.Set;

import de.bsvrz.dav.daf.main.ClientDavInterface;
import de.bsvrz.dav.daf.main.ClientReceiverInterface;
import de.bsvrz.dav.daf.main.Data;
import de.bsvrz.dav.daf.main.DataAndATGUsageInformation;
import de.bsvrz.dav.daf.main.DataDescription;
import de.bsvrz.dav.daf.main.DataState;
import de.bsvrz.dav.daf.main.ReceiveOptions;
import de.bsvrz.dav.daf.main.ReceiverRole;
import de.bsvrz.dav.daf.main.ResultData;
import de.bsvrz.dav.daf.main.Data.Array;
import de.bsvrz.dav.daf.main.Data.NumberValue;
import de.bsvrz.dav.daf.main.Data.ReferenceValue;
import de.bsvrz.dav.daf.main.Data.TextValue;
import de.bsvrz.dav.daf.main.Data.TimeValue;
import de.bsvrz.dav.daf.main.config.Aspect;
import de.bsvrz.dav.daf.main.config.AttributeGroup;
import de.bsvrz.dav.daf.main.config.AttributeGroupUsage;
import de.bsvrz.dav.daf.main.config.ConfigurationArea;
import de.bsvrz.dav.daf.main.config.ConfigurationChangeException;
import de.bsvrz.dav.daf.main.config.ConfigurationObject;
import de.bsvrz.dav.daf.main.config.DataModel;
import de.bsvrz.dav.daf.main.config.DynamicObjectType;
import de.bsvrz.dav.daf.main.config.MutableSet;
import de.bsvrz.dav.daf.main.config.SystemObject;
import de.bsvrz.dav.daf.main.config.SystemObjectType;
import de.bsvrz.sys.funclib.concurrent.UnboundedQueue;
import de.bsvrz.sys.funclib.debug.Debug;
import de.bsvrz.vew.sim.sim.EmpfaengerMitAnmeldung;
import de.bsvrz.vew.sim.sim.sender.Sender;
import de.bsvrz.vew.sim.sim.simulation.OfflineSimulation;
import de.bsvrz.vew.sim.sim.simulation.OnlineSimulation;
import de.bsvrz.vew.sim.sim.simulation.Simulation;
import de.bsvrz.vew.sim.sim.simulation.Zustand;
import de.bsvrz.vew.sim.sim.simulationsstrecke.SimulationsStrecke;
import de.bsvrz.vew.sim.sim.steuerung.SteuerungOfflineSimulation;

/**
 * Verwaltung von Simulationen. Mit einer Instanz dieser Klasse werden die dynamischen 
 * Simulations-Objekte (und die dazugehörigen Simulationsstrecken) eines Konfigurationsbereiches
 * verwaltet.<br>
 * Eine Instanz dieser Klasse empfängt und versendet Datensätze der logischen Schnittstelle
 * zum Verwalten (erzeugen und löschen) von Simulationen
 *
 * @version $Revision: 1.5 $ / $Date: 2009/11/17 15:37:48 $ / ($Author: Sans $)
 *
 * @author Dambach-Werke GmbH
 * @author Stefan Sans
 *
 */
public class VerwaltungSimulation implements EmpfaengerMitAnmeldung
{
  /**
   * Debug-Logger für Logging-Ausgaben
   */
  private final static Debug _debug = Debug.getLogger();

  /**
   * Verbindung zum Datenverteiler
   */
  private ClientDavInterface _connection;

  /**
   * Konfigurationsbereich, dessen dynamische Simulations-Objekte mit sieser Instanz
   * verwaltet werden
   */
  private ConfigurationArea _konfigBereich;

  /**
   * Konfigurationsobjekt zur Simulationsverwaltung
   */
  private ConfigurationObject _konfigObjekt;

  /**
   * Die verwalteten Simulationen
   */
  private Map<String, Simulation> _simulationen;

  /**
   * Alle Simulationsstrecken, aud die di verwalteten Simulationen verweisen
   */
  private Map<String, SimulationsStrecke> _simulationsStrecken;

  /**
   * Zugriff auf die verwalteten Simulationen.
   * @return nichtmodifizierbare <code>Map</code>, welche mit der PID als Schlüssel
   * den Zugriff auf die verwalteten Simulationen erlaubt
   */
  public Map<String, Simulation> getSimulationen()
  {
    return Collections.unmodifiableMap(_simulationen);
  }

  /**
   * Queue der zu bearbeitenden Verwaltungsanfragen
   */
  private UnboundedQueue<Data> _anfrageQueue = new UnboundedQueue<Data>();

  /**
   * Queue der zu versendenden Quittungen auf Verwaltungsanfragen
   */
  private UnboundedQueue<ResultData> _quittungQueue = new UnboundedQueue<ResultData>();

/*
  public Map<String, SimulationsStrecke> getSimulationsStrecken()
  {
    return Collections.unmodifiableMap(_simulationsStrecken);
  }
*/

  /**
   * Erzeugt eine neue Instanz von <code>VerwaltungSimulation</code>.
   * @param connection die Datenverteiler-Verbindung
   * @param konfigBereich der Konfigurationsbereich, aus dem die verwalteteten
   * dynamischen Simulations-Objekte gelesen werden bzw. in den neue Simulations-
   * Objekte geschrieben werden
   * @param konfigObjekt das Konfigurationsobjekt zum Verwalten von Simulationen
   * vom Typ typ.simulationsVerwaltung
   */
  public VerwaltungSimulation(ClientDavInterface connection,
      ConfigurationArea konfigBereich, ConfigurationObject konfigObjekt)
  {
    _connection = connection;
    _konfigBereich = konfigBereich;
    _konfigObjekt = konfigObjekt;

    _debug.config("Konfigurationsbereich: " + _konfigBereich);
    _debug.config("Konfigurationsobjekt: " + _konfigObjekt);

    _simulationen = new HashMap<String, Simulation>();
    _simulationsStrecken = new HashMap<String, SimulationsStrecke>();

    new QuittungSenden().start();
    new AnfrageBearbeiten().start();
  }

  /**
   * Bestimmt den Konfigurationsbereich für diese Simulationsverwaltung.
   * @return den Konfigurationsbereich für diese Simulationsverwaltung
   */
  public ConfigurationArea getKonfigBereich()
  {
    return _konfigBereich;
  }

  /**
   * Bestimmt das Datenmodell für diese Simulationsverwaltung.
   * @return das Datenmodell für diese Simulationsverwaltung
   */
  public DataModel getDatenModell()
  {
    return getConnection().getDataModel();
  }

  /**
   * Initialisiert die vorhandenen Simulationen (aus dem Konfigurationsbereich).
   * Die vorhandenen Simulationen werden in den Zustand "Stop" gebracht.
   */
  public void initialisiereVorhandeneSimulationen()
  {
    int mengeValid = 0;
    int alleValid = 0;
    Collection<SystemObject> menge = _konfigObjekt.getMutableSet("Simulationen").getElements();
    Collection<SystemObject> alle = getDatenModell().getType("typ.simulation").getObjects();

    for (SystemObject obj: alle)
    {
      if (!obj.isValid())
      {
        _debug.finer("Gelöschte Simulation " + obj + " ignoriert");
        continue;
      }

      SimulationKonfigDaten kd;
      try
      {
        kd = getKonfigDaten(obj);
      }
      catch(RuntimeException e)
      {
        _debug.error("Konfigurierende Daten konnten nicht ermmitelt werden für Simulation " +
            obj.getPid(), e);
        continue;
      }

      alleValid += 1;

      if (menge.contains(obj))
      {
        if (kd.ss.getSystemObject() != null && kd.ss.getSystemObject().isOfType("typ.simulationsStrecke"))
        {
          _debug.fine("Vorhandene Simulation " + obj + " verweist auf Simulationsstrecke " + kd.ss);
          if (SimulationsVarianten.getInstanz().isFrei(kd.sv))
          {
            _debug.fine("Vorhandene Simulation " + obj + " hat Simulationsvariante " + kd.sv);
            SimulationsVarianten.getInstanz().setBelegt(kd.sv);
            // Diese vorhande Simulation kann in den Zustand 'Neu' gebracht werden
            initialisiereSimulation(obj, Simulation.ZustandsBezeichner.Stop , true, true);
            mengeValid += 1;
          }
          else
          {
            /*
            _debug.error("Vorhandene Simulation " + obj + " hat ungültige Simulationsvariante " + kd._sv);
            loescheSimulation(obj);
            _debug.info("Vorhandene ungültige Simulation " + obj + " gelöscht");
            */
            _debug.error("Vorhandene Simulation " + obj + " wird NICHT verwendet wegen ungültiger Simulationsvariante " + kd.sv);
          }
        }
        else
        {
          /*
          _debug.error("Vorhandene Simulation " + obj + " verweist auf ungültige Simulationsstrecke " + kd._ss);
          loescheSimulation(obj);
          _debug.info("Vorhandene ungültige Simulation " + obj + " gelöscht");
          */
          _debug.error("Vorhandene Simulation " + obj + " wird NICHT verwendet wegen Verweis auf ungültige Simulationsstrecke " + kd.ss);
        }
      }
      else
      {
        SimulationsVarianten.getInstanz().setBelegt(kd.sv);
      }
    }

    /*
    for (SystemObject obj: alle)
    {
      if (!menge.contains(obj))
      {
        try
        {
          _debug.warning("loesche " + obj);
          obj.invalidate();
        }
        catch(ConfigurationChangeException e)
        {
          e.printStackTrace();
        }
      }
    }
    */

    _debug.info(mengeValid + " vorhandene Simulation(en) initialisiert aus der dyn. Menge von " +
        _konfigObjekt.getPid() + " - von ingesamt " + alleValid + " global gültige Simulationen");
  }

    public SystemObject getObjekt()
  {
    return _konfigObjekt;
  }

    public ClientDavInterface getConnection()
  {
    return _connection;
  }

    public void anmelden()
  {
    ReceiveOptions anmeldeoptionen = ReceiveOptions.normal();
    ReceiverRole empfaengerrolle = ReceiverRole.drain();

    AttributeGroup atg = getDatenModell().getAttributeGroup("atg.simulationVerwaltungAnfrage");
    Aspect asp = getDatenModell().getAspect("asp.anfrage");
    short simulationsVariante = 0;

    DataDescription datenbeschreibung = new DataDescription(atg, asp, simulationsVariante);

    getConnection().subscribeReceiver(this, getObjekt(), datenbeschreibung, anmeldeoptionen, empfaengerrolle);
  }

  public void update(ResultData[] resultData)
  {

    // Ausgaben
    for (ResultData data : resultData)
    {
      _debug.finest("Daten empfangen:\n" + data);

      Aspect asp = data.getDataDescription().getAspect();
      AttributeGroup atg = data.getDataDescription().getAttributeGroup();
      short sv = data.getDataDescription().getSimulationVariant();

      _debug.finer("Daten empfangen mit " + atg.getPid() + "/" +
          asp.getPid() + "/" + sv  + ":\n" + data);

      if (sv == 0 && asp.getPid().equals("asp.anfrage") &&
          atg.getPid().equals("atg.simulationVerwaltungAnfrage"))
      {
        boolean anmeldungGueltig = true;

        if (data.getDataState() == DataState.DATA)
        {
          _debug.fine("Bearbeite Daten mit " + atg.getPid() + "/" +
              asp.getPid() + "/" + sv + ":\n" + data);

          _anfrageQueue.put(data.getData());
        }
        else if (data.getDataState() == DataState.INVALID_SUBSCRIPTION)
        {
          anmeldungGueltig = false;

          _debug.error("Die Empfangs-Anmeldung auf " + data.getObject() +
              " mit " + atg + "/" + asp + "/" + sv + " ist ungültig");
        }
        if (_anmeldungVewAnfrGueltig == null)
        {
          synchronized(this)
          {
            _anmeldungVewAnfrGueltig = anmeldungGueltig ? Boolean.TRUE : Boolean.FALSE;
            notify();
          }
        }
        else
        {
          _anmeldungVewAnfrGueltig = anmeldungGueltig ? Boolean.TRUE : Boolean.FALSE;
        }
      }
    }
  }

  /**
   * Ist die Empfangsanmledung vollzogen und gültig oder ungültig weil bereits
   * andere Senke angemeldet ist?
   */
  private Boolean _anmeldungVewAnfrGueltig = null;

  /**
   * Wartet maximal timeout ms auf eine Empfangsanmeldung.
   * @param timeout die Zeit im ms, die max. gewartet werde soll
   * @return <code>true</code>, wenn innerhalb des Timeouts eine gültige oder keine Empfangsanmledung
   * erfolgte, <code>false</code> bei ungültiger Empfangsanmeldung innerhalb timeout
   */
  public boolean warteAnmeldungVewAnfrGueltig(long timeout)
  {
    if (_anmeldungVewAnfrGueltig == null)
    {
      if (timeout > 0)
      {
        _debug.config("Warte max. " + timeout +
            "ms auf Fehlermeldung für Empfangs-Anmeldung Verwaltungsanfrage");

        synchronized(this)
        {
          try
          {
            wait(timeout);
          }
          catch(InterruptedException e)
          {
          }
        }
      }
    }
    return _anmeldungVewAnfrGueltig == null ? true :_anmeldungVewAnfrGueltig.booleanValue();
  }

  /**
   * Bereinig für die Verwaltung von dyn. Simulations-Objekten bereitgestellten Ressourcen
   * @param simulation das dyn. Simulations-Objekt, für das bereinigt werden soll
   */
  protected void bereinigeSimulation(SystemObject simulation)
  {
    String pid = simulation.getPid();
    Simulation sim = _simulationen.remove(pid);
    if (sim != null)
    {
      sim.abmelden();
      sim.getSteuerung().sendeBeenden();
      _debug.fine("Simulation " + simulation + " bereinigt");
    }
    else
    {
      _debug.warning(simulation + " wird nicht als Simulation verwaltet");
    }
  }

  /**
   * Initialisiert diese Instanz zur Verwaltung eines dyn. Simulations-Objekts
   * @param simulation simulation das dyn. Simulations-Objekt, für das initialisiert werden soll
   * @param anfangsZustand Bezeichner für den Anfangszustand, in den die fertig initialisierte
   * Simulation gebracht wird
   * @param sendeBeenden soll "sendeBeenden" an die SWE StartStopp gesendet werden?
   * @param zustandPublizieren soll der finale Zustand publiziert werden?
   */
  protected void initialisiereSimulation(SystemObject simulation,
      Simulation.ZustandsBezeichner anfangsZustand, boolean sendeBeenden,
      boolean zustandPublizieren)
  {
    final Simulation sim;
    SimulationKonfigDaten kv = this.getKonfigDaten(simulation);

    // Simulationsstrecke
    String ssPid = kv.ss.getSystemObject().getPid();
    SimulationsStrecke ss = _simulationsStrecken.get(ssPid);
    if (ss == null)
    {
      ss = new SimulationsStrecke(kv.ss.getSystemObject(), getConnection());
      _simulationsStrecken.put(ssPid, ss);
      ss.anmelden();
    }

    // Simulation
    if (simulation.isOfType("typ.onlineSimulation"))
    {
      sim = new OnlineSimulation(simulation,kv.sv, ss, getConnection(), anfangsZustand);
    }
    else if (simulation.isOfType("typ.offlineSimulation"))
    {
      sim = new OfflineSimulation(simulation,kv.sv, ss, getConnection(), anfangsZustand);
    }
    else
    {
      throw new IllegalArgumentException("Objekt " + simulation +
          " mit falschem Typ " + simulation.getType() +
          " - Initialisierung fehlgeschlagen");
    }

    _simulationen.put(simulation.getPid(), sim);
    sim.anmelden();
    if (sendeBeenden)
    {
      sim.getSteuerung().sendeBeenden();
    }

    if (zustandPublizieren)
    {
      sim.getSteuerung().sendeZustand();
    }

    Runtime.getRuntime().addShutdownHook(new Thread()
    {
      public void run()
      {
        if (_simulationen.values().contains(sim))
        {
          // Falls diese Simulation noch verwaltet wird
          _debug.info("Aufräumen Simulation" + sim);
          Zustand zustand = sim.getZustand();
          if (zustand.getName().equals(Simulation.ZustandsBezeichner.Start.text) ||
              zustand.getName().equals(Simulation.ZustandsBezeichner.Pause.text))
          {
            sim.setZustand(Simulation.ZustandsBezeichner.Stop.text);
          }
          sim.getSteuerung().sendeBeenden();
        }
      }
    });
  }

  /**
   * Erzeugt ein neues dynmaisches Simulations-Objekt
   * @param name Name
   * @param simulationsVariante Simulationsvariante
   * @param simulationsStrecke Referenz auf Simulationsstrecke
   * @param typ PID des Typs der zu erzeugenden Simulation
   * @return das erzeugte dynamische Simulations-Objekt
   * @throws Exception Exception
   */
  protected SystemObject erzeugeSimulation(String name, short simulationsVariante,
      ReferenceValue simulationsStrecke, String typ)
    throws Exception
  {
    SystemObjectType sot = getDatenModell().getType(typ);

    if (!(sot instanceof DynamicObjectType) ||
        !sot.inheritsFrom(getDatenModell().getType("typ.simulation")))
    {
      throw new IllegalArgumentException("Der übergebene Typ \"" + typ + "\" ist ungültig");
    }

    AttributeGroup atg = getDatenModell().getAttributeGroup("atg.simulationsEigenschaften");
    Aspect asp = getDatenModell().getAspect("asp.eigenschaften");
    String pid = "simulation." + new DecimalFormat("000").format(simulationsVariante);

    SystemObject simulation = null;

    try
    {
      Data data = getConnection().createData(atg);
      data.getItem("SimulationsVariante").asUnscaledValue().set(simulationsVariante);
      data.getItem("SimulationsStreckenReferenz").asReferenceValue().setSystemObject(simulationsStrecke.getSystemObject());

      AttributeGroupUsage atgu = atg.getAttributeGroupUsage(asp);
      DataAndATGUsageInformation daaui = new DataAndATGUsageInformation(atgu, data);
      Collection<DataAndATGUsageInformation> col = new ArrayList<DataAndATGUsageInformation>();
      col.add(daaui);

      simulation = _konfigBereich.createDynamicObject((DynamicObjectType)sot, pid, name, col);
      _konfigObjekt.getMutableSet("Simulationen").add(simulation);
    }
    catch(Exception e)
    {
      if (simulation != null)
      {
        simulation.invalidate();
      }
      throw(e);
    }

    return simulation;
  }

  /**
   * Löscht ein dynamisches Simulations-Objekt
   * @param simulation das zu löschende dynamische Simulations-Objekt
   * @throws ConfigurationChangeException Exception
   */
  protected void loescheSimulation(SystemObject simulation)
    throws ConfigurationChangeException
  {
    boolean ausMengeGeloescht = false;
    MutableSet set = null;

    try
    {
      set = _konfigObjekt.getMutableSet("Simulationen");
      if (set.getElements().contains(simulation))
      {
        set.remove(simulation);
        ausMengeGeloescht = true;
      }
      simulation.invalidate();
    }
    catch (ConfigurationChangeException e)
    {
      if (ausMengeGeloescht)
      {
        // Stelle Ursprungszustand wieder her
        set.add(simulation);
        throw(e);
      }
    }
  }

  /**
   * Bestimmt die (relevaten) konfigurierenden Daten eines dyn. Simulations-Objekts
   * @param simulation das dyn. Simulations-Objekt, dessen konfigurierende Daten
   * bestimmt werden sollen
   * @return Instanz von <code>SimulationKonfigDaten</code> mit den gewünschten Daten
   */
  protected SimulationKonfigDaten getKonfigDaten(SystemObject simulation)
  {
    Aspect asp = getDatenModell().getAspect("asp.eigenschaften");
    AttributeGroup atg = getDatenModell().getAttributeGroup("atg.simulationsEigenschaften");
    Data data = simulation.getConfigurationData(atg, asp);

    short sv  = data.getItem("SimulationsVariante").asUnscaledValue().shortValue();
    ReferenceValue ss = data.getItem("SimulationsStreckenReferenz").asReferenceValue();

    SimulationKonfigDaten result = new SimulationKonfigDaten(sv, ss);

    return result;
  }

  /**
   * Hilfsklasse, modelliert die relevanten konfigurierenden Daten einer Simulation
   * @author Stefan Sans
   */
  protected static class SimulationKonfigDaten
  {
    /**
     * Erzeugt neue Instanz von <code>SimulationKonfigDaten</code>
     * @param sv die Simulationsvariante
     * @param ss Referenz auf die Simulatiopnsstrecke
     */
    SimulationKonfigDaten(short sv, ReferenceValue ss)
    {
      this.sv = sv;
      this.ss = ss;
    }

    /**
     * Simulationsvariante
     */
    final short sv;

    /**
     * Simulationsstrecke
     */
    final ReferenceValue ss; //

  }

  /**
   * Senden einer Antwort auf eine Simulations-Verwaltungs-Anfrage
   * @param absender das Stellverteterobjekt des Absenders dieser Antwort
   * - typischerweise das lokale Applikationsobjekt
   * @param absenderZeichen frei wählbare (User-) Absender-Id des Absenders
   * @param erzeugeErgebnis Ergebnis der Anfrage zum Erzeugen von Simulationen
   * @param loescheErgebnis Ergebnis der Anfrage zum Löschen von Simulationen
   */
  private void sendeAntwort(SystemObject absender, String absenderZeichen,
      ErzeugeErgebnis[] erzeugeErgebnis, LoescheErgebnis[] loescheErgebnis)
  {
    AttributeGroup atg = getDatenModell().getAttributeGroup("atg.simulationVerwaltungAntwort");
    Aspect asp = getDatenModell().getAspect("asp.antwort");
    short sv = 0;
    Data data = getConnection().createData(atg);

    TextValue textValue = data.getTextValue("AbsenderZeichen");
    textValue.setText(absenderZeichen);

    Array array = data.getArray("ErzeugeSimulationAntwort");
    array.setLength(erzeugeErgebnis.length);
    for (int i = 0; i < erzeugeErgebnis.length; i += 1)
    {
      Data datensatz = array.getItem(i);
      ReferenceValue simRef = datensatz.getReferenceValue("ReferenzErzeugteSimulation");
      simRef.setSystemObject(erzeugeErgebnis[i].simulation);
      NumberValue nv = datensatz.getUnscaledValue("ErzeugeSimulationZustand");
      nv.set(erzeugeErgebnis[i].zustand);
      nv = datensatz.getUnscaledValue("ErzeugeSimulationFehler");
      nv.set(erzeugeErgebnis[i].fehler);
    }

    array = data.getArray("LoescheSimulationAntwort");
    array.setLength(loescheErgebnis.length);
    for (int i = 0; i < loescheErgebnis.length; i += 1)
    {
      Data datensatz = array.getItem(i);
      NumberValue nv = datensatz.getUnscaledValue("LoescheSimulationZustand");
      nv.set(loescheErgebnis[i].zustand);
      nv = datensatz.getUnscaledValue("LoescheSimulationFehler");
      nv.set(loescheErgebnis[i].fehler);
    }

    _debug.fine("Sende Antwort an: " + absender);
    _quittungQueue.put(new ResultData(absender, new DataDescription(atg, asp, sv),
        getConnection().getTime(), data));
  }

  /**
   * Behandeln eine Simulations-Verwaltunjgs-Anfrage zum Erzeugen und/oder Löschen von
   * Simulationen
   * @param data Der Datensatz, der für diese Anfrage vom Datenverteiler empfangen wurde
   */
  private void behandleAnfrage(Data data)
  {
    SystemObject absender = data.getItem("AbsenderId").asReferenceValue().getSystemObject();
    String absenderZeichen = data.getItem("AbsenderZeichen").asTextValue().getText();

    Array array = data.getArray("ErzeugeSimulationAnfrage");
    ErzeugeErgebnis[] erzeugeErgebnis = new ErzeugeErgebnis[array.getLength()];
    for (int i = 0; i < array.getLength(); i += 1)
    {
      Data datenSatz = array.getItem(i);
      String typ = datenSatz.getItem("SimulationsArt").asTextValue().getText();
      if (typ.equals("Online"))
      {
        typ = "typ.onlineSimulation";
      }
      else if (typ.equals("Offline"))
      {
        typ = "typ.offlineSimulation";
      }
      else
      {
        typ = null;
      }

      if (typ != null)
      {
        String name = datenSatz.getItem("Name").asTextValue().getText();
        ReferenceValue ss = datenSatz.getItem("ReferenzSimulationsStrecke").asReferenceValue();
        if (ss.getSystemObject() != null && ss.getSystemObject().isOfType("typ.simulationsStrecke"))
        {
          synchronized (SimulationsVarianten.getInstanz())
          {
            short sv = SimulationsVarianten.getInstanz().getNiedrigsteFreie();
            SimulationsVarianten.getInstanz().setBelegt(sv);
            if (sv > 0 && sv < 1000)
            {
              SystemObject simulation;

              try
              {
                simulation = erzeugeSimulation(name, sv, ss, typ);
              }
              catch(Exception e)
              {
                _debug.error(e.getMessage());
                erzeugeErgebnis[i] = new ErzeugeErgebnis(null,
                    ErzeugeErgebnis.ZUSTAND_NICHT_ERZEUGT, ErzeugeErgebnis.FEHLER_NICHT_ERZEUGT);
                continue;
              }
              initialisiereSimulation(simulation, Simulation.ZustandsBezeichner.neu,
                  false, true);
              if (typ.equals("typ.offlineSimulation"))
              {
                Array zeitbereich = datenSatz.getArray("OfflineSimulationsZeitbereich");
                if (zeitbereich.getLength() != 0)
                {
                  OfflineSimulation sim = (OfflineSimulation)_simulationen.get(simulation.getPid());
                  SteuerungOfflineSimulation steuerung = sim.getSteuerung();
                  Data beginnEnde = zeitbereich.getItem(0);
                  TimeValue timeValue = beginnEnde.getTimeValue("SimulationsBeginn");
                  steuerung.setBeginn(timeValue.getMillis());
                  timeValue = beginnEnde.getTimeValue("SimulationsEnde");
                  steuerung.setEnde(timeValue.getMillis());
                  steuerung.sendeZeitbereich();
                }
                else
                {
                  _debug.warning("Offline-Simulation " + simulation +
                      " ohne initialen Zeitbereich erstellt");
                }
              }
              erzeugeErgebnis[i] = new ErzeugeErgebnis(simulation,
                  ErzeugeErgebnis.ZUSTAND_ERZEUGT, ErzeugeErgebnis.KEIN_FEHLER);
            }
            else
            {
              _debug.error("Keine freie Simulationsvariante");
              erzeugeErgebnis[i] = new ErzeugeErgebnis(null,
                  ErzeugeErgebnis.ZUSTAND_NICHT_ERZEUGT, ErzeugeErgebnis.FEHLER_SIM_VARIANTE);
            }
          }
        }
        else
        {
          _debug.error("Simulationsstreckenreferenz ungültig");
          erzeugeErgebnis[i] = new ErzeugeErgebnis(null,
              ErzeugeErgebnis.ZUSTAND_NICHT_ERZEUGT, ErzeugeErgebnis.FEHLER_SIM_STRECKE);
        }
      }
      else
      {
        _debug.error("Unzulässiger Simulationstyp");
        erzeugeErgebnis[i] = new ErzeugeErgebnis(null,
            ErzeugeErgebnis.ZUSTAND_NICHT_ERZEUGT, ErzeugeErgebnis.FEHLER_NICHT_ERZEUGT);
      }
    }

    array = data.getArray("LoescheSimulationAnfrage");
    LoescheErgebnis[] loescheErgebnis = new LoescheErgebnis[array.getLength()];

    for (int i = 0; i < array.getLength(); i += 1)
    {
      Data datenSatz = array.getItem(i);
      ReferenceValue ref = datenSatz.getItem("ReferenzZuLoeschendeSimulation").asReferenceValue();
      if (ref.getSystemObject() != null && ref.getSystemObject().isOfType("typ.simulation"))
      {
        String pid = ref.getSystemObject().getPid();
        Simulation sim = _simulationen.get(pid);

        if (sim != null)
        {
          /* Physikalisches Löschen des dynamischen Simulationsobjektes
           * ist möglich aus dem Zustand 'Gelöscht'.
           */
          if (sim.getZustand().getName().equals(Simulation.ZustandsBezeichner.geloescht.text))
          {
            /*
             * loescheErgebnis wird hier vorläufig auf ZUSTAND_GELOESCHT/KEIN_FEHLER gesetzt
             * Die Statusüberprüfung sowie das eigentliche Löschen erfolgt in AnfragePostProzess
             */
            loescheErgebnis[i] = new LoescheErgebnis(ref.getSystemObject(),
                LoescheErgebnis.ZUSTAND_GELOESCHT, LoescheErgebnis.KEIN_FEHLER);
          }
          else
          {
            _debug.error("Simulation " + ref.getSystemObject() + " muss sich im Zustand \"" +
                Simulation.ZustandsBezeichner.geloescht + "\" befinden");
            loescheErgebnis[i] = new LoescheErgebnis(ref.getSystemObject(),
                LoescheErgebnis.ZUSTAND_NICHT_GELOESCHT, LoescheErgebnis.FEHLER_FALSCHER_ZUSTAND);
          }
        }
        else
        {
          _debug.error("Simulation " + ref.getSystemObject() + " nicht gefunden");
          loescheErgebnis[i] = new LoescheErgebnis(ref.getSystemObject(),
              LoescheErgebnis.ZUSTAND_NICHT_GELOESCHT, LoescheErgebnis.FEHLER_NICHT_GEFUNDEN);
        }
      }
      else
      {
        _debug.error("Objekt " + ref.getSystemObject() + " ist als zu löschende Simulation ungültig");
        loescheErgebnis[i] = new LoescheErgebnis(ref.getSystemObject(),
            LoescheErgebnis.ZUSTAND_NICHT_GELOESCHT, LoescheErgebnis.FEHLER_NICHT_GELOESCHT);
      }
    }

    new AnfragePostProzess(absender, absenderZeichen, erzeugeErgebnis, loescheErgebnis).start();
  }

  /**
   * Post-Prozessing für Verwaltungsanfragen- es wird eine Quittung für die Verwaltungsanfrage generiert.
   * Wenn Simulationsobjekte zum Löschen anstehen, wird der Status von Konfiguration, Parametrierung und
   * Archiv ermittelt und ggf. auf einen positiven Status gewartet - dann erst wird gelöscht und die
   * komplette Quittung versendet.
   *
   * @author Dambach Werke
   */
  class AnfragePostProzess extends Thread implements ClientReceiverInterface
  {
    private SystemObject _absender;
    private String _absenderZeichen;
    private ErzeugeErgebnis[] _erzeugeErgebnis;
    private LoescheErgebnis[] _loescheErgebnis;

    private Set<SystemObject> _angemeldeteObjekte = new LinkedHashSet<SystemObject>();

    private DataDescription _ddKonfiguration;
    private DataDescription _ddParametrierung;
    private DataDescription _ddArchiv;

    private Set<SystemObject> _freigabeKonfiguration = new HashSet<SystemObject>();
    private Set<SystemObject> _freigabeParametrierung = new HashSet<SystemObject>();
    private Set<SystemObject> _freigabeArchiv = new HashSet<SystemObject>();

    public final static long TIMEOUT = 180000;

    /**
     * Erzeugt eine neue Instanz vom Typ <code>AnfragePostProzess</code> für die übergebenen Daten einer
     * Verwaltungsanfrage.
     *
     * @param absender Absender-Id der Verwaltungsanfrage
     * @param absenderZeichen Absender-Zeichen der Verwaltungsanfrage
     * @param erzeugeErgebnis Ergebnis der zu erzeugenden Simulationsobjekte
     * @param loescheErgebnis Ergebnis (vorläufig) der zu löschenden Simulationsobjekte
     */
    public AnfragePostProzess(SystemObject absender, String absenderZeichen,
        ErzeugeErgebnis[] erzeugeErgebnis, LoescheErgebnis[] loescheErgebnis)
    {
      _absender = absender;
      _absenderZeichen = absenderZeichen;
      _erzeugeErgebnis = erzeugeErgebnis;
      _loescheErgebnis = loescheErgebnis;

      Aspect asp = getConnection().getDataModel().getAspect("asp.zustand");
      AttributeGroup atg = getConnection().getDataModel().getAttributeGroup("atg.simulationsStatusKonfiguration");
      _ddKonfiguration = new DataDescription(atg, asp, (short)0);
      atg = getConnection().getDataModel().getAttributeGroup("atg.simulationsStatusParametrierung");
      _ddParametrierung = new DataDescription(atg, asp, (short)0);
      atg = getConnection().getDataModel().getAttributeGroup("atg.simulationsStatusArchiv");
      _ddArchiv = new DataDescription(atg, asp, (short)0);
    }

    @Override
    public void run()
    {
      for (LoescheErgebnis result: _loescheErgebnis)
      {
        if (result.zustand == LoescheErgebnis.ZUSTAND_GELOESCHT &&
            result.fehler == LoescheErgebnis.KEIN_FEHLER)
        {
          Simulation simulation = _simulationen.get(result.simulation.getPid());
          if (simulation != null)
          {
            if (simulation.getStatusKonfiguration() ==  Simulation.KONFIG_STATUS_GELOESCHT)
            {
              _freigabeKonfiguration.add(result.simulation);
            }
            if (simulation.getStatusParametrierung() ==  Simulation.PARAM_STATUS_NICHT_AKTIV)
            {
              _freigabeParametrierung.add(result.simulation);
            }
            if (simulation.getStatusArchiv() ==  Simulation.ARC_STATUS_NICHT_AKTIV)
            {
              _freigabeArchiv.add(result.simulation);
            }
          }
          _angemeldeteObjekte.add(result.simulation);
        }
      }

      if (!_angemeldeteObjekte.isEmpty())
      {
        getConnection().subscribeReceiver(this, _angemeldeteObjekte, _ddKonfiguration,
            ReceiveOptions.normal(), ReceiverRole.receiver());
        getConnection().subscribeReceiver(this, _angemeldeteObjekte, _ddParametrierung,
            ReceiveOptions.normal(), ReceiverRole.receiver());
        getConnection().subscribeReceiver(this, _angemeldeteObjekte, _ddArchiv,
            ReceiveOptions.normal(), ReceiverRole.receiver());

        _debug.fine("Warte " + (TIMEOUT > 0 ? "max. " + TIMEOUT/1000 + "s " : "") +
            "auf Freigabe durch Konfiguration, Parametrierung, Archiv");
        try
        {
          synchronized(this)
          {
            wait(TIMEOUT);
          }
        }
        catch(InterruptedException e)
        {
          _debug.warning("Thread unerwartet unterbrochen");
        }
      }

      if (_angemeldeteObjekte.size() > 0)
      {
        _debug.config(_angemeldeteObjekte.size() + " dynamische(s) Simulationsobjekt(e)konnte nicht gelöscht werden," +
            "\nda dafür von Konfiguration, Parametrierung oder Archiv keine Löschfreigabe vorliegt");
      }

      for (int i = 0; i < _loescheErgebnis.length; i += 1)
      {
        if (_loescheErgebnis[i].zustand == LoescheErgebnis.ZUSTAND_GELOESCHT &&
            _loescheErgebnis[i].fehler == LoescheErgebnis.KEIN_FEHLER)
        {
          if (_angemeldeteObjekte.contains(_loescheErgebnis[i].simulation))
          {
            _loescheErgebnis[i] = new LoescheErgebnis(_loescheErgebnis[i].simulation,
                LoescheErgebnis.ZUSTAND_NICHT_GELOESCHT, LoescheErgebnis.FEHLER_FALSCHER_ZUSTAND);

            getConnection().unsubscribeReceiver(this, _loescheErgebnis[i].simulation, _ddKonfiguration);
            getConnection().unsubscribeReceiver(this, _loescheErgebnis[i].simulation, _ddParametrierung);
            try
            {
              getConnection().unsubscribeReceiver(this, _loescheErgebnis[i].simulation, _ddArchiv);
            }
            catch (IllegalStateException e)
            {
              _debug.fine("IllegalStateException beim abmelden als Empfänger von __ddArchiv ignoriert");
            }
          }
        }
      }

      sendeAntwort(_absender, _absenderZeichen, _erzeugeErgebnis, _loescheErgebnis);
    }

    public void update(ResultData[] resultData)
    {
      for (ResultData data: resultData)
      {
        if (data.getDataDescription().equals(_ddKonfiguration))
        {
          if (!_freigabeKonfiguration.contains(data.getObject()) && data.hasData() &&
              data.getData().getItem("KonfigurationZustand").asTextValue().getText().equals("gelöscht"))
          {
            _freigabeKonfiguration.add(data.getObject());
          }
          if (_freigabeKonfiguration.contains(data.getObject()))
          {
            _debug.fine("Freigabe von der Konfiguration zum Löschen von " + data.getObject().getPid());
            bearbeiteDaten(data.getObject());
          }
          else
          {
            _debug.config("(Noch) KEINE Freigabe von der Konfiguration zum Löschen von " +
                data.getObject().getPid());
          }
        }
        else if (data.getDataDescription().equals(_ddParametrierung))
        {
          if (!_freigabeParametrierung.contains(data.getObject()) && data.hasData() &&
              data.getData().getItem("ParametrierungBereit").asTextValue().getText().equals("Nein"))
          {
            _freigabeParametrierung.add(data.getObject());
          }
          if (_freigabeParametrierung.contains(data.getObject()))
          {
            _debug.fine("Freigabe von der Parametrierung zum Löschen von " + data.getObject().getPid());
            bearbeiteDaten(data.getObject());
          }
          else
          {
            _debug.config("(Noch) KEINE Freigabe von der Parametrierung zum Löschen von " +
                data.getObject().getPid());
          }
        }
        else if (data.getDataDescription().equals(_ddArchiv))
        {
          if (!_freigabeArchiv.contains(data.getObject()) && data.hasData() &&
              data.getData().getItem("ArchivBereit").asTextValue().getText().equals("Nein"))
          {
            _freigabeArchiv.add(data.getObject());
          }
          if (_freigabeArchiv.contains(data.getObject()))
          {
            _debug.fine("Freigabe vom Archiv zum Löschen von " + data.getObject().getPid());
            bearbeiteDaten(data.getObject());
          }
          else
          {
            _debug.config("(Noch) KEINE Freigabe vom Archiv zum Löschen von " +
                data.getObject().getPid());
          }
        }
      }
    }

    private void bearbeiteDaten(SystemObject objekt)
    {
      if (_freigabeKonfiguration.contains(objekt) &&
          _freigabeParametrierung.contains(objekt) &&
          _freigabeArchiv.contains(objekt))
      {
        _freigabeKonfiguration.remove(objekt);
        _freigabeParametrierung.remove(objekt);
        _freigabeArchiv.remove(objekt);
        if (_angemeldeteObjekte.remove(objekt))
        {
          getConnection().unsubscribeReceiver(this, objekt, _ddKonfiguration);
          getConnection().unsubscribeReceiver(this, objekt, _ddParametrierung);
          try
          {
            getConnection().unsubscribeReceiver(this, objekt, _ddArchiv);
          }
          catch (IllegalStateException e)
          {
            _debug.fine("IllegalStateException beim abmelden als Empfänger von __ddArchiv ignoriert");
          }
        }

        boolean fehler = false;
        bereinigeSimulation(objekt);
        try
        {
          loescheSimulation(objekt);
        }
        catch(Exception e)
        {
          fehler = true;
          _debug.error("Dynamisches Simulationsobjekt " + objekt.getPid() + " konnte nicht gelöscht werden", e);
          for (int i = 0; i < _loescheErgebnis.length;  i+= 1)
          {
            if (_loescheErgebnis[i].simulation == objekt)
            {
              _loescheErgebnis[i] = new LoescheErgebnis(objekt,
                  LoescheErgebnis.ZUSTAND_NICHT_GELOESCHT, LoescheErgebnis.FEHLER_NICHT_GEFUNDEN);
              break;
            }
          }
          initialisiereSimulation(objekt, Simulation.ZustandsBezeichner.geloescht, false, false);
        }

        if (!fehler)
        {
          SimulationsVarianten.getInstanz().setFrei(getKonfigDaten(objekt).sv);
          _debug.config("Dynamisches Simulationsobjekt " + objekt.getPid() + " erfolgreich gelöscht");
        }
      }

      if (_angemeldeteObjekte.isEmpty())
      {
        synchronized(this)
        {
          notify();
        }
      }
    }
  }

  /**
   * Modelliert das Ergebnis einer (einzelnen) Anfrage zum erzeugen einer Simulation
   * @author Stefan Sans
   */
  private static class ErzeugeErgebnis
  {
    public static final int ZUSTAND_ERZEUGT = 0;
    public static final int ZUSTAND_NICHT_ERZEUGT = 1;
    public static final int KEIN_FEHLER = 0;
    public static final int FEHLER_SIM_VARIANTE = 1;
    public static final int FEHLER_SIM_STRECKE = 2;
    public static final int FEHLER_NICHT_ERZEUGT = 3;

    public final SystemObject simulation;
    public final int zustand;
    public final int fehler;

    ErzeugeErgebnis(SystemObject simulation, int zustand, int fehler)
    {
      this.simulation = simulation;
      this.zustand = zustand;
      this.fehler = fehler;
    }
  }

  /**
   * Modelliert das Ergebnis einer (einzelnen) Anfrage zum löschen einer Simulation
   * @author Stefan Sans
   */
  private static class LoescheErgebnis
  {
    public static final int ZUSTAND_GELOESCHT = 0;
    public static final int ZUSTAND_NICHT_GELOESCHT = 1;
    public static final int KEIN_FEHLER = 0;
    public static final int FEHLER_FALSCHER_ZUSTAND = 1;
    public static final int FEHLER_NICHT_GEFUNDEN = 2;
    public static final int FEHLER_NICHT_GELOESCHT = 3;

    public final SystemObject simulation;
    public final int zustand;
    public final int fehler;

    LoescheErgebnis(SystemObject simulation, int zustand, int fehler)
    {
      this.simulation = simulation;
      this.zustand = zustand;
      this.fehler = fehler;
    }
  }

  /**
   * Thread, der die Queue der zu bearbeitenden Verwaltungsanfragen abarbeitet.
   * Hier werden die empfangenen Vewaltungsanfragen zur Bearbeitung weitergegeben
   *
   * @author Dambach Werke
   */
  class AnfrageBearbeiten extends Thread
  {
    /**
     * Instanziiert einen neuen Thread "Anfrage-Bearbeitungs-Thread"
     *
     */
    AnfrageBearbeiten()
    {
      super();

      setName("Anfrage-Bearbeitungs-Thread");
    }

    @Override
    public void start()
    {
      super.start();
      _debug.config(this + " gestartet");
    }

    @Override
    public void run()
    {
      // Abarbeitung wartender Verwaltungsanfragen
      while (true)
      {
        try
        {
          Data data = _anfrageQueue.take();
          if (data != null)
          {
            behandleAnfrage(data);
          }
        }
        catch(InterruptedException e)
        {
          _debug.warning("Thread wurde unerwartet unterbrochen", e);
        }
      }
    }
  }

  /**
   * Thread, der die Queue der zu versendenden Quittungen auf Verwaltungsanfragen abarbeitet.
   * Hier werden die warteten Quittungen zur tatsächlichen Sendung weitergegeben
   * 
   * @author Dambach Werke
   */
  class QuittungSenden extends Thread
  {
    /**
     * Instanziiert einen neuen Thread "Quittungs-Sende-Thread"
     *
     */
    QuittungSenden()
    {
      super();
      
      setName("Quittungs-Sende-Thread");
    }
    
    @Override
    public void start()
    {
      super.start();
      _debug.config(this + " gestartet");
    }
    
    @Override
    public void run()
    {
      // Abarbeitung wartender zu sendender Quittungen auf Verwaltungsanfragen
      while (true)
      {
        try
        {
          ResultData resultData = _quittungQueue.take();
          if (resultData != null)
          {
            Sender.send(getConnection(), resultData);
          }
        }
        catch(InterruptedException e)
        {
          _debug.warning("Thread wurde unerwartet unterbrochen", e);
        }
      }
    }
  }

}
